@@ -64,6 +64,7 @@
 
 static volatile int stop_reading = 0;
 static volatile int stop_writing = 0;
+static volatile int pause_writing = 0;
 static volatile int ring_in = 0;
 static volatile int ring_out = 0;
 static int cmdfd;
@@ -77,7 +78,7 @@ static sbc_t sbc;
 static int raw = 0;
 static int rtcfd = -1;
 static volatile int packetticks;
-static pthread_t pid;
+static pthread_t transmit_pid, listen_pid;
 
 // number of RTC ticks per second... a power of 2 up to 8192
 #define TICKS 8192
@@ -133,6 +134,16 @@ static ssize_t __write(int fd, const void *buf, size_t count)
 }
 #endif
 
+#if 0
+static void dump_packet(void *p, int size)
+{
+	uint8_t *c = (uint8_t *) p;
+	while (size-- > 0)
+		fprintf(stderr, " %02x\n", *c++);
+	fprintf(stderr, "\n");
+}
+#endif
+
 // Detect when the user presses CTRL+C
 static void sig_term(int sig)
 {
@@ -150,9 +161,12 @@ static void send_packet()
 {
 	int idx, written;
 
+	//printf("sending (%d != %d)\n", ring_in, ring_out);
+	
 	// no data to send
 	if(ring_in == ring_out) return;
 
+	
 	// find the next packet
 	// skip to the last encoded packet if we're live to reduce latency
 	do {
@@ -182,7 +196,20 @@ static void *transmit_thread(void *init_data)
 	unsigned long data;
 	int ticks = packetticks;
 
+	// Time reference
+	static struct timeval staticcounter = {0,0};
+	static int icount = 0;
+
 	while(!stop_writing) {
+
+		static struct timeval timediff = {0,0};
+		
+	if(staticcounter.tv_sec==0)
+	{
+		gettimeofday(&staticcounter, NULL);
+		gettimeofday(&timediff, NULL);
+	}
+	
 		// blocks until next rtc event
 		if(read(rtcfd, &data, sizeof(data)) < sizeof(data)) {
 			perror("/dev/rtc read");
@@ -190,15 +217,33 @@ static void *transmit_thread(void *init_data)
 			break;
 		}
 
-		// more than one click may have passed
-		// the api docs for /dev/rtc are extremely incomplete
-		// status is in the low byte
-		// number of ticks since last read is in 3 high bytes
-		ticks -= data >> 8;
+	struct timeval timeofday, duration;//, playtime, theoricaldate;
+	gettimeofday(&timeofday, NULL);
+	timersub(&timeofday, &staticcounter, &duration);
 
-		while(ticks <= 0) {
-			send_packet();
-			ticks += packetticks;
+	// Display data once per second
+	icount++;
+	if(duration.tv_sec>0)
+	{
+		printf("read %lX %d times / sec.\n", data, icount);
+		gettimeofday(&staticcounter, NULL);
+		icount=0;
+	}
+	
+		if(pause_writing) {
+			sleep(1);
+		} else {
+			// more than one click may have passed
+			// the api docs for /dev/rtc are extremely incomplete
+			// status is in the low byte
+			// number of ticks since last read is in 3 high bytes
+			ticks -= data >> 8;
+
+//			while(ticks <= 0)
+			{
+				send_packet();
+				ticks += packetticks;
+			}
 		}
 	}
 	pthread_exit(NULL);
@@ -207,7 +252,7 @@ static void *transmit_thread(void *init_data)
 
 static int compute_ticks(unsigned long sleeptime)
 {
-	return 1.0 * ((TICKS * sleeptime)/(1000000));
+	return ((1.0 * TICKS * sleeptime)/(1000000));
 }
 
 // Usage
@@ -218,10 +263,11 @@ static void usage()
 	fprintf(stderr, "filename is an audio file or device, or stdin if not specified.\n\n");
 
 	fprintf(stderr, "Options:\n");
-	fprintf(stderr, " -h print this usage and exit\n");
+	fprintf(stderr, " -h print this usage and exit\n\n");
 	//fprintf(stderr, " -f fork and run as a daemon\n");
-	fprintf(stderr, " -n send packets as fast as they are encoded instead of timing them\n");
-	fprintf(stderr, " -p use pthreads and rtc for packet timing\n");
+	fprintf(stderr, " -i use setitimer for packet timing (deprecated)\n");
+	fprintf(stderr, " -n send packets as fast as they are encoded instead of timing them (deprecated)\n");
+	fprintf(stderr, " -p use pthreads timer (default)\n\n");
 	fprintf(stderr, " -t thrifty (repeat for very thrifty)\n");
 	fprintf(stderr, " -v print verbose output\n\n");
 
@@ -392,16 +438,6 @@ static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t
 	return sk;
 }
 
-#if 0
-static void dump_packet(void *p, int size)
-{
-	uint8_t *c = (uint8_t *) p;
-	while (size-- > 0)
-		fprintf(stderr, " %02x\n", *c++);
-	fprintf(stderr, "\n");
-}
-#endif
-
 // Prepare packet headers
 static void init_request(struct avdtp_header * header, int request_id)
 {
@@ -418,6 +454,50 @@ static void init_request(struct avdtp_header * header, int request_id)
 	transaction = (transaction + 1) & 0xf;
 }
 
+static void init_response(struct avdtp_header * header, int response_type)
+{
+	// leave signal_id and transaction label since we are reusing the request
+	header->packet_type = PACKET_TYPE_SINGLE;
+	header->message_type = response_type;
+
+	// clear rfa bits
+	header->rfa0 = 0;
+}
+
+// monitor the control connection for pause/play signals from headset
+// note this signaling is in the avdtp core; avrcp signaling is different
+static void *listen_thread(void *init_data) 
+{
+	int size;
+	struct stream_cmd cmd;
+	printf("Listen thread running\n");
+	while(!stop_writing) {
+		size = read(cmdfd, &cmd, sizeof(cmd));
+		if(size == sizeof(cmd)) {
+			printf("received signal %d from set\n", cmd.header.signal_id);
+			if(cmd.header.signal_id == AVDTP_SUSPEND) {
+				pause_writing = 1;
+			} else if(cmd.header.signal_id == AVDTP_START) {
+				pause_writing = 0;
+			} else {
+				printf("unexpected headset directive %d\n", cmd.header.signal_id);
+			}
+			// ack the command regardless
+			// take a shortcut and reuse the command struct (knock one byte off length)
+			init_response(&cmd.header, MESSAGE_TYPE_ACCEPT);
+			if (write(cmdfd, &cmd, sizeof(cmd)-1) != sizeof(cmd)-1) {
+				fprintf(stderr, "Couldn't ack %d\n", cmd.header.signal_id);
+			}
+		}
+		else
+		{
+			printf("error while receiving %d\n", size);
+		}
+	}
+	pthread_exit(NULL);
+	return NULL;
+}
+
 // Analyse the SEIDs the sink has sent to us
 static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned short *psm)
 {
@@ -427,7 +507,7 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 	struct getcap_resp cap_resp;
 	struct set_config s_config;
 	struct set_config_resp s_resp;
-	struct open_stream_cmd open_stream;
+	struct stream_cmd open_stream;
 	struct open_stream_rsp open_resp;
 	fprintf(stderr, "SEID = %d\n", seid);
 
@@ -440,13 +520,26 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		return (-1);
 	}
 	else fprintf(stderr, "Requested Capabilities for SEID = %d\n",seid);
-	if (read(s, &cap_resp, sizeof(cap_resp)) < sizeof(cap_resp) ||
+	if (read(s, &cap_resp, sizeof(cap_resp)) < sizeof(cap_resp) || 
 			cap_resp.header.message_type == MESSAGE_TYPE_REJECT ||
-			cap_resp.media_type != AUDIO_MEDIA_TYPE ||
-			cap_resp.media_codec_type != SBC_MEDIA_CODEC_TYPE) {
+			cap_resp.media_type != AUDIO_MEDIA_TYPE) {
 		fprintf(stderr, "Didn't receive SBC codec parameters (first) for SEID = %d\n", seid);
 		return (-1);
 	}
+	if (cap_resp.media_codec_type != SBC_MEDIA_CODEC_TYPE) {
+		if(cap_resp.media_codec_type == MPEG12_MEDIA_CODEC_TYPE) {
+			fprintf(stderr, "mpeg decoder found: layer3=%d channel_mode=%x frequency=%x bitrate0=%x bitrate1=%x\n",
+				cap_resp.codec_elements.mpeg12_elements.layer3,
+				cap_resp.codec_elements.mpeg12_elements.channel_mode,
+				cap_resp.codec_elements.mpeg12_elements.frequency,
+				cap_resp.codec_elements.mpeg12_elements.bitrate0,
+				cap_resp.codec_elements.mpeg12_elements.bitrate1);
+		} else {
+			fprintf(stderr, "no SBC decoder here: media_codec_type is %d not %d for SEID = %d\n", 
+				cap_resp.media_codec_type, SBC_MEDIA_CODEC_TYPE, seid);
+		}
+		return (-1);
+	}
 
 	fprintf(stderr, "Got capabilities response\n");
 
@@ -469,7 +562,7 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 2;
 		break;
 	}
-	s_config.sbc_elements.channel_mode = v;
+	s_config.codec_elements.sbc_elements.channel_mode = v;
 
 	switch(sbc.rate) {
 	case 16000:
@@ -486,8 +579,8 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 2; 
 		break;
 	}
-	s_config.sbc_elements.frequency = v;
-	s_config.sbc_elements.allocation_method = 1 << 1;
+	s_config.codec_elements.sbc_elements.frequency = v;
+	s_config.codec_elements.sbc_elements.allocation_method = 1 << 1;
 
 	switch(sbc.subbands) {
 	case 4:
@@ -498,7 +591,7 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 1;
 		break;
 	}
-	s_config.sbc_elements.subbands = v;
+	s_config.codec_elements.sbc_elements.subbands = v;
 
 	switch(sbc.blocks) {
 	case 4:
@@ -515,31 +608,31 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 1;
 		break;
 	}
-	s_config.sbc_elements.block_length = v;
+	s_config.codec_elements.sbc_elements.block_length = v;
 /*
   int_seid was 1 before
 s_config was set from sbc_info before:
-        s_config.sbc_elements.channel_mode = 8 >> sbc_info.channel_mode;
-        s_config.sbc_elements.frequency = 8 >> sbc_info.sampling_frequency;
-        s_config.sbc_elements.allocation_method = 1 << sbc_info.allocation_method;
-        s_config.sbc_elements.subbands = 2 >> sbc_info.subbands;
-        s_config.sbc_elements.block_length = 8 >> sbc_info.blocks;
+        s_config.codec_elements.sbc_elements.channel_mode = 8 >> sbc_info.channel_mode;
+        s_config.codec_elements.sbc_elements.frequency = 8 >> sbc_info.sampling_frequency;
+        s_config.codec_elements.sbc_elements.allocation_method = 1 << sbc_info.allocation_method;
+        s_config.codec_elements.sbc_elements.subbands = 2 >> sbc_info.subbands;
+        s_config.codec_elements.sbc_elements.block_length = 8 >> sbc_info.blocks;
  */
-	s_config.sbc_elements.min_bitpool = cap_resp.sbc_elements.min_bitpool;
-	s_config.sbc_elements.max_bitpool = cap_resp.sbc_elements.max_bitpool;
+	s_config.codec_elements.sbc_elements.min_bitpool = cap_resp.codec_elements.sbc_elements.min_bitpool;
+	s_config.codec_elements.sbc_elements.max_bitpool = cap_resp.codec_elements.sbc_elements.max_bitpool;
 
-        //dump_packet(&s_config.sbc_elements, sizeof(s_config.sbc_elements)); //exit(0);
+        //dump_packet(&s_config.codec_elements.sbc_elements, sizeof(s_config.codec_elements.sbc_elements)); //exit(0);
 
-	if (!(cap_resp.sbc_elements.channel_mode & s_config.sbc_elements.channel_mode))
+	if (!(cap_resp.codec_elements.sbc_elements.channel_mode & s_config.codec_elements.sbc_elements.channel_mode))
 		fprintf(stderr, "headset does not support this channel mode\n");
 
-	if (!(cap_resp.sbc_elements.frequency & s_config.sbc_elements.frequency))
+	if (!(cap_resp.codec_elements.sbc_elements.frequency & s_config.codec_elements.sbc_elements.frequency))
 		fprintf(stderr, "headset does not support this frequency\n");
 
-	if (!(cap_resp.sbc_elements.allocation_method & s_config.sbc_elements.allocation_method))
+	if (!(cap_resp.codec_elements.sbc_elements.allocation_method & s_config.codec_elements.sbc_elements.allocation_method))
 		fprintf(stderr, "headset does not support this allocation_method\n");
 
-	if (!(cap_resp.sbc_elements.subbands & s_config.sbc_elements.subbands))
+	if (!(cap_resp.codec_elements.sbc_elements.subbands & s_config.codec_elements.sbc_elements.subbands))
 		fprintf(stderr, "headset does not support this subbands setting\n");
 
 	if (write(s, &s_config, sizeof(s_config)) != sizeof(s_config)) {
@@ -589,11 +682,11 @@ int main(int argc, char *argv[])
 	struct sigaction sa;
 	struct sepd_req put_req;
 	struct sepd_resp get_resp;
-	struct start_stream_cmd start_stream;
+	struct stream_cmd start_stream;
 	struct start_stream_rsp start_resp;
 	struct media_packet_header packet_header;
 	struct media_payload_header payload_header;
-	struct close_stream_cmd close_stream;
+	struct stream_cmd close_stream;
 	struct close_stream_rsp close_resp;
 	struct sbc_frame_header sbc_info;
 
@@ -611,7 +704,7 @@ int main(int argc, char *argv[])
 	int rate = 44100;
 	int channels = 2;
 	int little = 0;
-	int pthreads = 0;
+	int pthreads = 1;
 	int notimer = 0;
 	int seid, last_seid_index;
 	char *addrstr;
@@ -635,11 +728,14 @@ int main(int argc, char *argv[])
 	bacpy(&src, BDADDR_ANY);
 
 	// process command line
-	while((i = getopt(argc, argv, "tpnvfdmlsr:")) >= 0) {
+	while((i = getopt(argc, argv, "itpnvfdmlsr:")) >= 0) {
 		switch(i) {
 		case 't':
 			thrifty++;
 			break;
+		case 'i':
+			pthreads = 0;
+			break;
 		case 'p':
 			pthreads = 1;
 			break;
@@ -862,7 +958,7 @@ int main(int argc, char *argv[])
 
 	fprintf(stderr, "received %d capabilities\n", last_seid_index + 1);
 
-	for(i=0; i <= last_seid_index; i++) {
+	for(i=last_seid_index; i >= 0; i--) {
 		if (process_seid(cmdfd, &get_resp.infos[i], &psm_stream) == 0) {
 			seid = get_resp.infos[i].acp_seid;
 			break;
@@ -1034,7 +1130,8 @@ int main(int argc, char *argv[])
 						perror("rtc enable ioctl");
 						exit(1);
 					}
-					pthread_create(&pid, NULL, transmit_thread, NULL);
+					pthread_create(&transmit_pid, NULL, transmit_thread, NULL);
+					pthread_create(&listen_pid, NULL, listen_thread, NULL);
 				} else {
 					itimer.it_interval.tv_sec = itimer.it_value.tv_sec = 0;
 					itimer.it_interval.tv_usec = itimer.it_value.tv_usec = (sleeptime*(87))/100;
@@ -1069,24 +1166,28 @@ int main(int argc, char *argv[])
 
 	fprintf(stderr, "Sent %d packets\n", seq_num);
 
+	if(pthreads) {
+		fprintf(stderr, "joining x thread\n");
+		pthread_join(transmit_pid, &retval);
+		// have to stop the listener thread because it's blocked on i/o
+		fprintf(stderr, "stopping r thread\n");
+		pthread_kill(listen_pid, SIGINT);
+		fprintf(stderr, "closing rtc\n");
+		close(rtcfd);
+	}
+
 	// signal the stream close
 
 	memset(&close_stream, 0, sizeof(close_stream));
 	memset(&close_resp, 0, sizeof(close_resp));
 
-	if(pthreads) {
-		fprintf(stderr, "joining thread\n");
-		pthread_join(pid, &retval);
-		fprintf(stderr, "closing rtc\n");
-		close(rtcfd);
-	}
-
 	// the stream-close used to make the iTech headset lock up and require it to be powercycled
 	// should be tested again now that we drain the queue properly
 
 	init_request(&close_stream.header, AVDTP_CLOSE);
 	close_stream.acp_seid = seid;
 
+#if 1
 	if (write(cmdfd, &close_stream, sizeof(close_stream)) != sizeof(close_stream)) {
 		fprintf(stderr, "couldn't send close_stream\n");
 		close(streamfd);
@@ -1104,6 +1205,7 @@ int main(int argc, char *argv[])
 	}
 
 	fprintf(stderr, "Got close stream confirm\n");
+#endif
 
 	fprintf(stderr, "closing stream\n");
 	close(streamfd);
