@@ -43,32 +43,22 @@
 #include <alsa/pcm_external.h>
 #include <alsa/timer.h>
 
+#include <pthread.h>
+
 #include <netinet/in.h>
 
 #include "sbc.h"
 #include "../a2dp.h"
 
 #define NONSPECAUDIO 1
-time_t timestamp=0;
-uint16_t seq_num=1;
-int frame_count=0;
 #define BUFS 1024
-static char bufe[BUFS];
-int lenbufe=0;
-
-unsigned long nbytes=0;
-unsigned long total_time=0;
-struct timeval t0;
-struct itimerval itimer;
-struct sigaction sa;
-struct timeval tsend;
-int fpr=0;
-
 
 #define min(X, Y)  ((X) < (Y) ? (X) : (Y))
 #define DBG(fmt, arg...)  printf("DEBUG: %s: " fmt "\n" , __FUNCTION__ , ## arg)
 //#define DBG(D...)
 
+struct sigaction        actions;
+
 static void a2dp_init(void) __attribute__ ((constructor));
 static void a2dp_exit(void) __attribute__ ((destructor));
 
@@ -111,7 +101,7 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 	struct getcap_resp cap_resp;
 	struct set_config s_config;
 	struct set_config_resp s_resp;
-	struct open_stream_cmd open_stream;
+	struct stream_cmd open_stream;
 	struct open_stream_rsp open_resp;
 	DBG("SEID = %d", seid);
 
@@ -119,7 +109,8 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 	init_request(&put_req.header, AVDTP_GET_CAPABILITIES);
 	put_req.acp_seid = seid;
 
-	if (write(s, &put_req, sizeof(put_req)) != sizeof(put_req)) {
+	if (write(s, &put_req, sizeof(put_req)) != sizeof(put_req))
+	{
 		DBG("Couldn't request capabilities for SEID = %d", seid);
 		return (-1);
 	}
@@ -155,7 +146,7 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 2;
 		break;
 	}
-	s_config.sbc_elements.channel_mode = v;
+	s_config.codec_elements.sbc_elements.channel_mode = v;
 
 	switch(sbc->rate) {
 	case 16000:
@@ -172,8 +163,8 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 2; 
 		break;
 	}
-	s_config.sbc_elements.frequency = v;
-	s_config.sbc_elements.allocation_method = 1 << 1;
+	s_config.codec_elements.sbc_elements.frequency = v;
+	s_config.codec_elements.sbc_elements.allocation_method = 1 << 1;
 
 	switch(sbc->subbands) {
 	case 4:
@@ -184,7 +175,7 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 1;
 		break;
 	}
-	s_config.sbc_elements.subbands = v;
+	s_config.codec_elements.sbc_elements.subbands = v;
 
 	switch(sbc->blocks) {
 	case 4:
@@ -201,23 +192,23 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 		v = 1;
 		break;
 	}
-	s_config.sbc_elements.block_length = v;
-	s_config.sbc_elements.min_bitpool = cap_resp.sbc_elements.min_bitpool;
-	s_config.sbc_elements.max_bitpool = cap_resp.sbc_elements.max_bitpool;
+	s_config.codec_elements.sbc_elements.block_length = v;
+	s_config.codec_elements.sbc_elements.min_bitpool = cap_resp.codec_elements.sbc_elements.min_bitpool;
+	s_config.codec_elements.sbc_elements.max_bitpool = cap_resp.codec_elements.sbc_elements.max_bitpool;
 
-	if (!(cap_resp.sbc_elements.channel_mode & s_config.sbc_elements.channel_mode)) {
+	if (!(cap_resp.codec_elements.sbc_elements.channel_mode & s_config.codec_elements.sbc_elements.channel_mode)) {
 		DBG("headset does not support this channel mode");
 	}
 
-	if (!(cap_resp.sbc_elements.frequency & s_config.sbc_elements.frequency)) {
+	if (!(cap_resp.codec_elements.sbc_elements.frequency & s_config.codec_elements.sbc_elements.frequency)) {
 		DBG("headset does not support this frequency");
 	}
 
-	if (!(cap_resp.sbc_elements.allocation_method & s_config.sbc_elements.allocation_method)) {
+	if (!(cap_resp.codec_elements.sbc_elements.allocation_method & s_config.codec_elements.sbc_elements.allocation_method)) {
 		DBG("headset does not support this allocation_method");
 	}
 
-	if (!(cap_resp.sbc_elements.subbands & s_config.sbc_elements.subbands)) {
+	if (!(cap_resp.codec_elements.sbc_elements.subbands & s_config.codec_elements.sbc_elements.subbands)) {
 		DBG("headset does not support this subbands setting");
 	}
 
@@ -256,8 +247,8 @@ static int process_seid(int s, struct acp_seid_info * get_seid_resp, unsigned sh
 
 	*psm = 25;
 	return 0;
-
 }
+
 // Detect whether A2DP Sink is present at the destination or not
 static int detect_a2dp(bdaddr_t *src, bdaddr_t *dst, unsigned short *psm, unsigned long *flags)
 {
@@ -270,14 +261,17 @@ static int detect_a2dp(bdaddr_t *src, bdaddr_t *dst, unsigned short *psm, unsign
 	int tries;
 
 	tries = 0;
-	while(!(sess = sdp_connect(src, dst, SDP_RETRY_IF_BUSY))) {
+	while(!(sess = sdp_connect(src, dst, SDP_RETRY_IF_BUSY)))
+	{
 		DBG("retrying sdp connect: %s", strerror(errno));
 		sleep(1);
-		if(++tries > 10) {
+		if(++tries > 10)
+		{
 			break;
 		}
 	}
-	if (!sess) {
+	if (!sess)
+	{
 		DBG( "Warning: failed to connect to SDP server: %s", strerror(errno));
 		if(psm) *psm = 25;
 		if(flags) *flags = 0;
@@ -293,13 +287,15 @@ static int detect_a2dp(bdaddr_t *src, bdaddr_t *dst, unsigned short *psm, unsign
 	sdp_list_free(search, 0);
 	sdp_list_free(attrid, 0);
 
-	if (err) {
+	if (err)
+	{
 		DBG( "Service Search failed: %s", strerror(errno));
 		sdp_close(sess);
 		return -1;
 	}
 
-	for (; seq; seq = next) {
+	for (; seq; seq = next)
+	{
 		sdp_record_t *rec = (sdp_record_t *) seq->data;
 
 		DBG( "Found A2DP Sink");
@@ -325,7 +321,8 @@ static int detect_a2dp(bdaddr_t *src, bdaddr_t *dst, unsigned short *psm, unsign
 	if (flags)
 		*flags = 0;
 
-	for (; seq; seq = next) {
+	for (; seq; seq = next)
+	{
 		sdp_record_t *rec = (sdp_record_t *) seq->data;
 		uint16_t vendor, product, version;
 
@@ -355,7 +352,9 @@ done:
 	sdp_close(sess);
 	return 0;
 }
-// Connecting on PSM 25
+
+// open an L2CAP connection on psm to the headset at dst
+// called with psm 25 for the command connection and later with a discovered psm for the stream
 static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t *mtu)
 {
 	struct sockaddr_l2 addr;
@@ -366,24 +365,21 @@ static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t
 
 	sk = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
 	if (sk < 0) {
-		DBG( "Can't create socket. %s(%d)",
-			strerror(errno), errno);
+		DBG( "Can't create socket. (errno=%d:%s)", errno, strerror(errno));
 		return -1;
 	}
 	memset(&addr, 0, sizeof(addr));
 	addr.l2_family = AF_BLUETOOTH;
 	bacpy(&addr.l2_bdaddr, src);
 	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		DBG( "Can't bind socket. %s(%d)",
-						strerror(errno), errno);
+		DBG( "Can't bind socket. (errno=%d:%s)", errno, strerror(errno));
 		return -1;
 	}
 
 	/* Get default options */
 	opt = sizeof(opts);
 	if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, &opt) < 0) {
-		DBG( "Can't get default L2CAP options. %s(%d)",
-						strerror(errno), errno);
+		DBG( "Can't get default L2CAP options. (errno=%d:%s)", errno, strerror(errno));
 		return -1;
 	}
 
@@ -393,8 +389,7 @@ static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t
 		//opts.imtu = *mtu;
 	}
 	if (setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, opt) < 0) {
-		DBG( "Can't set L2CAP options. %s(%d)",
-						strerror(errno), errno);
+		DBG( "Can't set L2CAP options. (errno=%d:%s)", errno, strerror(errno));
 		return -1;
 	}
 
@@ -405,8 +400,8 @@ static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t
 
 	tries = 0;
 	while (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		DBG("Can't connect to %s on psm %d. %s(%d)",
-				batostr(&addr.l2_bdaddr), psm, strerror(errno), errno);
+		DBG("Can't connect to %s on psm %d. (errono=%d:%s)",
+				batostr(&addr.l2_bdaddr), psm, errno, strerror(errno));
 		sleep(1);
 		++tries;
 		if(++tries > 10) {
@@ -416,8 +411,7 @@ static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t
 	}
 	opt = sizeof(opts);
 	if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &opts, &opt) < 0) {
-		DBG( "Can't get L2CAP options. %s(%d)",
-						strerror(errno), errno);
+		DBG( "Can't get L2CAP options. (errno=%d:%s)", errno, strerror(errno));
 		close(sk);
 		return -1;
 	}
@@ -431,11 +425,11 @@ static int do_connect(bdaddr_t *src, bdaddr_t *dst, unsigned short psm, uint16_t
 	return sk;
 }
 
-static int connect_stream(bdaddr_t *src, bdaddr_t *dst, int *cmdfd_return, sbc_t *sbc) {
+static int connect_stream(bdaddr_t *src, bdaddr_t *dst, int *cmdfd_return, sbc_t *sbc, int* seid_return) {
 	int cmdfd;
 	struct getcap_req put_req;
 	struct sepd_resp get_resp;
-	struct start_stream_cmd start_stream;
+	struct stream_cmd start_stream;
 	struct start_stream_rsp start_resp;
 	int seid, last_seid_index;
 	int size;
@@ -484,6 +478,11 @@ static int connect_stream(bdaddr_t *src, bdaddr_t *dst, int *cmdfd_return, sbc_t
 			break;
 		}
 	}
+	if (size == 2 && get_resp.header.message_type == 0 && get_resp.header.packet_type == 0) {
+		// Try again if this fails once (BlueAnt headset)
+		DBG("Discovery Response is the Wrong Size, Trying Read Again");
+		size = read(cmdfd, &get_resp, sizeof(get_resp));
+	}
 	if (size < sizeof(get_resp) - MAX_ADDITIONAL_CODEC_OCTETS) {
 		DBG("couldn't get avdtp_discover");
 		close(cmdfd);
@@ -541,8 +540,9 @@ static int connect_stream(bdaddr_t *src, bdaddr_t *dst, int *cmdfd_return, sbc_t
 	}
 
 	DBG("Got start stream confirm");
-
+	
 	*cmdfd_return = cmdfd;
+	*seid_return = seid;
 	return streamfd;
 }
 
@@ -555,12 +555,29 @@ typedef struct snd_pcm_a2dp {
 	bdaddr_t dst;
 	int sk;
 	int control_sk;
+	int seid;
 	sbc_t sbc;
 	snd_pcm_sframes_t num;
 	unsigned char buf[1024];
 	unsigned int len;
 	unsigned int frame_bytes;
 	int use_rfcomm;
+
+	char bufe[BUFS];
+	int lenbufe;//=0;
+
+	unsigned long nbytes;//=0;
+
+	struct timeval tsend;
+
+	time_t timestamp;//=0;
+	uint16_t seq_num;//=1;
+	int frame_count;//=0;
+
+	// Used to control stream from headset
+	int stop_writing;// = 0;
+	int pause_writing;// = 0;
+	pthread_t hListenThread;
 } snd_pcm_a2dp_t;
 
 static void inline a2dp_get(snd_pcm_a2dp_t *a2dp)
@@ -594,8 +611,15 @@ static int a2dp_stop(snd_pcm_ioplug_t *io)
 
 	DBG("a2dp %p", a2dp);
 
+	DBG("Listen thread terminating");
+	a2dp->stop_writing = 1;
+	pthread_kill(a2dp->hListenThread, SIGALRM);
+	pthread_join(a2dp->hListenThread, NULL);
+	DBG("Listen thread terminated");
+
 	a2dp->len = 0;
 
+	DBG("OK");
 	return 0;
 }
 
@@ -606,6 +630,73 @@ static snd_pcm_sframes_t a2dp_pointer(snd_pcm_ioplug_t *io)
 	return a2dp->num;
 }
 
+static void init_response(struct avdtp_header * header, int response_type)
+{
+	// leave signal_id and transaction label since we are reusing the request
+	header->packet_type = PACKET_TYPE_SINGLE;
+	header->message_type = response_type;
+
+	// clear rfa bits
+	header->rfa0 = 0;
+}
+
+// monitor the control connection for pause/play signals from headset
+// note this signaling is in the avdtp core; avrcp signaling is different
+static void *listen_thread(void * param) 
+{
+	snd_pcm_a2dp_t* a2dp = (snd_pcm_a2dp_t*)param;
+
+	printf("Listen thread running\n");
+
+	// Set a timeout to close thread
+	struct timeval t = { 5, 0 };
+	setsockopt( a2dp->control_sk, SOL_SOCKET, SO_SNDTIMEO, &t, sizeof(t));
+	setsockopt( a2dp->control_sk, SOL_SOCKET, SO_RCVTIMEO, &t, sizeof(t));
+
+	// Loop until end of writing
+	while(!a2dp->stop_writing)
+	{
+		int size;
+		struct stream_cmd cmd;
+		printf("Awaiting command\n");
+		size = read(a2dp->control_sk, &cmd, sizeof(cmd));
+		if(size == sizeof(cmd))
+		{
+			printf("Received signal %d from set\n", cmd.header.signal_id);
+			if(cmd.header.signal_id == AVDTP_SUSPEND)
+			{
+				a2dp->pause_writing = 1;
+			}
+			else if(cmd.header.signal_id == AVDTP_START)
+			{
+				a2dp->pause_writing = 0;
+			}
+			else
+			{
+				printf("Unexpected headset directive %d\n", cmd.header.signal_id);
+			}
+			// ack the command regardless
+			// take a shortcut and reuse the command struct (knock one byte off length)
+			init_response(&cmd.header, MESSAGE_TYPE_ACCEPT);
+			if (write(a2dp->control_sk, &cmd, sizeof(cmd)-1) != sizeof(cmd)-1)
+			{
+				fprintf(stderr, "Couldn't ack %d\n", cmd.header.signal_id);
+			}
+		}
+		else
+		{
+			if(errno!=EAGAIN)
+				printf("Error while receiving %d (errno=%d:%s)\n", size, errno, strerror(errno));
+			// Got end signal, leave
+			if(errno==EINTR)
+				break;
+		}
+	}
+
+	printf("%s: ending perfeclty\n", __FUNCTION__);
+	return NULL;
+}
+
 static void sleeptill(struct timeval *t, struct timeval *dt)
 {
 	struct timeval tc,dtc;
@@ -627,6 +718,7 @@ static void sleeptill(struct timeval *t, struct timeval *dt)
 	return;
 }
 
+/*
 // returns time to wait ie difference between tsend and current time
 // if time has come, advances tsend
 static int time_to_wait(struct timeval *tsend, struct timeval *dt)
@@ -649,7 +741,6 @@ static int time_to_wait(struct timeval *tsend, struct timeval *dt)
 	return dtc.tv_usec;
 }
 
-
 // transfers around correct time postions
 static snd_pcm_sframes_t a2dp_transfer(snd_pcm_ioplug_t *io,
 			const snd_pcm_channel_area_t *areas,
@@ -668,22 +759,22 @@ static snd_pcm_sframes_t a2dp_transfer(snd_pcm_ioplug_t *io,
 	codesize=a2dp->sbc.subbands*a2dp->sbc.blocks*a2dp->sbc.channels*2; // size of data encoded by sbc_encode in one call
 	datatoread=min(codesize,size*a2dp->frame_bytes); // amount of data to read
 	buf = (char *) areas->addr + (areas->first + areas->step * offset) / 8;
-    	if(lenbufe<codesize && lenbufe+datatoread<sizeof(bufe)){ // if not enough data in bufe to encode and there is space in bufe
-		memcpy(bufe+lenbufe,buf,datatoread);// we read data to bufe
-		lenbufe+=datatoread;
+    	if(a2dp->lenbufe<codesize && a2dp->lenbufe+datatoread<sizeof(a2dp->bufe)){ // if not enough data in bufe to encode and there is space in bufe
+		memcpy(a2dp->bufe+a2dp->lenbufe,buf,datatoread);// we read data to bufe
+		a2dp->lenbufe+=datatoread;
 	}
 	else{datatoread=0;}//nothing has been read
 
-	if(lenbufe>=codesize && a2dp->len + a2dp->sbc.len < 678){ // if enough data in bufe to encode and not enough frame to fill up mtu: encoding
-		change_endian(bufe,codesize); // changing the endianness
-		len = sbc_encode(&(a2dp->sbc), bufe, codesize); //encode
-		memmove(bufe, bufe + len, lenbufe - len); //shift the bufe                                 
-		lenbufe-=len;
-		nbytes+=len;
+	if(a2dp->lenbufe>=codesize && a2dp->len + a2dp->sbc.len < 678){ // if enough data in bufe to encode and not enough frame to fill up mtu: encoding
+		change_endian(a2dp->bufe,codesize); // changing the endianness
+		len = sbc_encode(&(a2dp->sbc), a2dp->bufe, codesize); //encode
+		memmove(a2dp->bufe, a2dp->bufe + len, a2dp->lenbufe - len); //shift the bufe                                 
+		a2dp->lenbufe-=len;
+		a2dp->nbytes+=len;
 		sleeptime += a2dp->sbc.duration;
 		if (len <= 0)
 			return len;
-		frame_count++;
+		a2dp->frame_count++;
 		memcpy(a2dp->buf + a2dp->len, a2dp->sbc.data, a2dp->sbc.len); // copy encoded frames into a2dp->buf
 		a2dp->len+=a2dp->sbc.len;
 		if (a2dp->state == BT_CONNECTED)
@@ -692,30 +783,30 @@ static snd_pcm_sframes_t a2dp_transfer(snd_pcm_ioplug_t *io,
 	}		
 
 	if(a2dp->len + a2dp->sbc.len > 678){ // if packet is formed
-		dt.tv_usec=1000000*a2dp->sbc.subbands*a2dp->sbc.blocks*frame_count/io->rate; // time interval between transmitions
+		dt.tv_usec=1000000*a2dp->sbc.subbands*a2dp->sbc.blocks*a2dp->frame_count/io->rate; // time interval between transmitions
 		dt.tv_sec=0;
-		if(time_to_wait(&tsend, &dt)==0){ // time to send data
+		if(time_to_wait(&a2dp->tsend, &dt)==0){ // time to send data
 			memset(&payload_header, 0, sizeof(payload_header)); // fill up the headers
 			memset(&packet_header, 0, sizeof(packet_header)); //---
-			payload_header.frame_count=frame_count;
+			payload_header.frame_count=a2dp->frame_count;
 			packet_header.v = 2;
 			packet_header.pt = 1;
-			packet_header.sequence_number = htons(seq_num);
-			packet_header.timestamp = htonl(timestamp);
+			packet_header.sequence_number = htons(a2dp->seq_num);
+			packet_header.timestamp = htonl(a2dp->timestamp);
 			packet_header.ssrc = htonl(1);
-			timestamp += (a2dp->sbc.blocks + 1)*4 * (a2dp->sbc.subbands + 1)*4;
+			a2dp->timestamp += (a2dp->sbc.blocks + 1)*4 * (a2dp->sbc.subbands + 1)*4;
 			memcpy(a2dp->buf, &packet_header, sizeof(packet_header)); //copy the headers to buf
 			memcpy(a2dp->buf + sizeof(packet_header), &payload_header, sizeof(payload_header));//---
 			write(a2dp->sk,a2dp->buf,a2dp->len); // sending the packet
 			a2dp->len = sizeof(packet_header)+sizeof(payload_header); //inital position in buf, just after headers
-			frame_count=0;
+			a2dp->frame_count=0;
 			sleeptime=0;
-			seq_num++;
+			a2dp->seq_num++;
 		}else{usleep(1);}
 	}
 	return datatoread / a2dp->frame_bytes;
 }
-
+*/
 // also works but sleeps between transfers
 static snd_pcm_sframes_t a2dp_transfer2(snd_pcm_ioplug_t *io,
 			const snd_pcm_channel_area_t *areas,
@@ -727,55 +818,95 @@ static snd_pcm_sframes_t a2dp_transfer2(snd_pcm_ioplug_t *io,
 	struct media_packet_header packet_header;
 	struct media_payload_header payload_header;
 	int codesize,datatoread;
-	unsigned long sleeptime;
+	unsigned long sleeptime=0;
 	int written;
-	
 	struct timeval dt;
-	
+	struct timeval timeofday;
+	gettimeofday(&timeofday, NULL);
+
+	// Time reference
+	static struct timeval staticcounter = {0,0};
+	static int sbc_bps=0;
+	if(staticcounter.tv_sec==0)
+	{
+		gettimeofday(&staticcounter, NULL);
+	}
+
+	struct timeval duration;
+	timersub(&timeofday, &staticcounter, &duration);
+
+	// Display bandwidth once per second
+	if(duration.tv_sec>0)
+	{
+		printf("SBC Bandwidth = %d kbps [%d bps]\n", sbc_bps/1024, sbc_bps);
+		// Reset static counter
+		gettimeofday(&staticcounter, NULL);
+		sbc_bps=0;
+	}
 
 	codesize=a2dp->sbc.subbands*a2dp->sbc.blocks*a2dp->sbc.channels*2;
 	datatoread=min(codesize,size*a2dp->frame_bytes);
+
+	/*
+	DBG("(%2d%3d): %d + %d * %d (=%d)",
+	    (timeofday.tv_sec      )%100,  // s
+	    (timeofday.tv_usec/1000)%1000, // ms
+	    (int)areas->first, (int)areas->step, offset, (areas->first + areas->step * offset));
+
+	DBG("(%2d%3d): %d * %d * %d * 2 (=%d) x %d",
+	    (timeofday.tv_sec      )%100,  // s
+	    (timeofday.tv_usec/1000)%1000, // ms
+	    a2dp->sbc.subbands, // 8
+	    a2dp->sbc.blocks, // 16
+	    a2dp->sbc.channels, // 2
+	    codesize, // 512
+	    datatoread); // 512
+	*/
 	buf = (char *) areas->addr + (areas->first + areas->step * offset) / 8;
-    	if(lenbufe<codesize){
-		memcpy(bufe+lenbufe,buf,datatoread);
-		lenbufe+=datatoread;
+   	if(a2dp->lenbufe<codesize)
+	{
+		memcpy(a2dp->bufe+a2dp->lenbufe,buf,datatoread);
+		a2dp->lenbufe+=datatoread;
 	}  
 	else{datatoread=0;}
 
-	if(lenbufe>=codesize){ //enough data to encode
-		change_endian(bufe,codesize); // changing the endianness
-		len = sbc_encode(&(a2dp->sbc), bufe, codesize); //encode
-		memmove(bufe, bufe + len, lenbufe - len); //shift the bufe                                 
-		lenbufe-=len;
-		nbytes+=len;
+	if(a2dp->lenbufe>=codesize)
+	{
+		//enough data to encode
+		change_endian(a2dp->bufe,codesize); // changing the endianness
+		len = sbc_encode(&(a2dp->sbc), a2dp->bufe, codesize); //encode
+		memmove(a2dp->bufe, a2dp->bufe + len, a2dp->lenbufe - len); //shift the bufe                                 
+		a2dp->lenbufe-=len;
+		a2dp->nbytes+=len;
 		sleeptime += a2dp->sbc.duration;
 		if (len <= 0)
 			return len;
 		if(a2dp->len + a2dp->sbc.len > 678)	{ // time to prepare and send the packet
 			dt.tv_sec=0;
-			dt.tv_usec=1000000*a2dp->sbc.subbands*a2dp->sbc.blocks*frame_count/io->rate;
+			dt.tv_usec=1000000*a2dp->sbc.subbands*a2dp->sbc.blocks*a2dp->frame_count/io->rate;
 			memset(&payload_header, 0, sizeof(payload_header));
 			memset(&packet_header, 0, sizeof(packet_header));
-			payload_header.frame_count=frame_count;
+			payload_header.frame_count=a2dp->frame_count;
 			packet_header.v = 2;
 			packet_header.pt = 1;
-			packet_header.sequence_number = htons(seq_num);
-			packet_header.timestamp = htonl(timestamp);
+			packet_header.sequence_number = htons(a2dp->seq_num);
+			packet_header.timestamp = htonl(a2dp->timestamp);
 			packet_header.ssrc = htonl(1);
-			timestamp += (a2dp->sbc.blocks + 1)*4 * (a2dp->sbc.subbands + 1)*4;
+			a2dp->timestamp += (a2dp->sbc.blocks + 1)*4 * (a2dp->sbc.subbands + 1)*4;
 			memcpy(a2dp->buf, &packet_header, sizeof(packet_header));
 			memcpy(a2dp->buf + sizeof(packet_header), &payload_header, sizeof(payload_header));
-			sleeptill(&tsend, &dt);
+			sleeptill(&a2dp->tsend, &dt);
 			if((written = write(a2dp->sk,a2dp->buf,a2dp->len)) != a2dp->len) {
 				DBG("Wrote %d not %d bytes; errno %s(%d)", written, a2dp->len,
 					strerror(errno), errno);
 			}
+		        sbc_bps += written;
 			a2dp->len = sizeof(packet_header)+sizeof(payload_header);
-			frame_count=0;
+			a2dp->frame_count=0;
 			sleeptime=0;
-			seq_num++;
+			a2dp->seq_num++;
 		}
-		frame_count++;
+		a2dp->frame_count++;
 		memcpy(a2dp->buf + a2dp->len, a2dp->sbc.data, a2dp->sbc.len);
 		a2dp->len+=a2dp->sbc.len;
 		if (a2dp->state == BT_CONNECTED)
@@ -787,12 +918,31 @@ static snd_pcm_sframes_t a2dp_transfer2(snd_pcm_ioplug_t *io,
 static int a2dp_close(snd_pcm_ioplug_t *io)
 {
 	snd_pcm_a2dp_t *a2dp = io->private_data;
+	struct stream_cmd close_stream;
+	struct close_stream_rsp close_resp;
 
-	DBG("a2dp %p", a2dp);
+	DBG("a2dp Destroying %p", a2dp);
+	memset(&close_stream, 0, sizeof(close_stream));
+	memset(&close_resp, 0, sizeof(close_resp));
+
+	// the stream-close used to make the iTech headset lock up and require it to be powercycled
+	// should be tested again now that we drain the queue properly
+
+	init_request(&close_stream.header, AVDTP_CLOSE);
+	close_stream.acp_seid = a2dp->seid;
+	if (write(a2dp->control_sk, &close_stream, sizeof(close_stream)) != sizeof(close_stream))
+	{
+		fprintf(stderr, "Couldn't send close_stream (errno=%d:%s)\n", errno, strerror(errno));
+		close(a2dp->control_sk);
+		a2dp->control_sk = -1;
+		close(a2dp->sk);
+		a2dp->sk = -1;
+	}
 
 	a2dp->len = 0;
 
 	a2dp_put(a2dp);
+	DBG("OK");
 
 	return 0;
 }
@@ -829,9 +979,6 @@ static int a2dp_prepare(snd_pcm_ioplug_t *io)
 
 	a2dp->sbc.rate = io->rate;
 	a2dp->sbc.channels = io->channels;
-	a2dp->sbc.subbands = 8; // safe default
-	a2dp->sbc.blocks = 16; // safe default
-	a2dp->sbc.bitpool = 32; // recommended value 53, safe default is 32
 
 	return 0;
 }
@@ -913,10 +1060,12 @@ static int a2dp_connect(snd_pcm_a2dp_t *a2dp)
 	socklen_t len;
 	int sk;
 	int control_sk = -1;
+	DBG("a2dp %p (sk=%d, control_sk=%d)", a2dp, a2dp->sk, a2dp->control_sk);
 
-	DBG("a2dp %p", a2dp);
+	a2dp->seid = -1;
 
-	if(a2dp->use_rfcomm) {
+	if(a2dp->use_rfcomm)
+	{
 		sk = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
 		if (sk < 0)
 			return -errno;
@@ -925,7 +1074,8 @@ static int a2dp_connect(snd_pcm_a2dp_t *a2dp)
 		addr.rc_family = AF_BLUETOOTH;
 		bacpy(&addr.rc_bdaddr, &a2dp->src);
 		
-		if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+		if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0)
+		{
 			close(sk);
 			return -errno;
 		}
@@ -935,7 +1085,8 @@ static int a2dp_connect(snd_pcm_a2dp_t *a2dp)
 		bacpy(&addr.rc_bdaddr, &a2dp->dst);
 		addr.rc_channel = 1;
 		
-		if (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+		if (connect(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0)
+		{
 			close(sk);
 			return -errno;
 		}
@@ -943,7 +1094,8 @@ static int a2dp_connect(snd_pcm_a2dp_t *a2dp)
 		memset(&addr, 0, sizeof(addr));
 		len = sizeof(addr);
 		
-		if (getsockname(sk, (struct sockaddr *) &addr, &len) < 0) {
+		if (getsockname(sk, (struct sockaddr *) &addr, &len) < 0)
+		{
 			close(sk);
 			return -errno;
 		}
@@ -951,13 +1103,19 @@ static int a2dp_connect(snd_pcm_a2dp_t *a2dp)
 		bacpy(&a2dp->src, &addr.rc_bdaddr);
 
 		fcntl(sk, F_SETFL, fcntl(sk, F_GETFL) | O_NONBLOCK);
-	} else {
-		sk = connect_stream(&a2dp->src, &a2dp->dst, &control_sk, &a2dp->sbc);
+	}
+	else
+	{
+		sk = connect_stream(&a2dp->src, &a2dp->dst, &control_sk, &a2dp->sbc, &a2dp->seid);
 	}
 
 	a2dp->sk = sk;
 	a2dp->control_sk = control_sk;
 
+	// Start listen thread
+	a2dp->pause_writing = 0;
+	a2dp->stop_writing = 0;
+	pthread_create(&a2dp->hListenThread, NULL, listen_thread, (void*)a2dp);
 	return 0;
 }
 
@@ -971,7 +1129,7 @@ static int a2dp_constraint(snd_pcm_a2dp_t *a2dp)
 	unsigned int format[2], channel[2], rate[2];
 	int err;
 
-	DBG("a2dp %p", a2dp);
+	DBG("TEST a2dp %p", a2dp);
 
 	err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_ACCESS, 2, access_list);
 	if (err < 0)
@@ -1040,6 +1198,7 @@ static inline snd_pcm_a2dp_t *a2dp_alloc(void)
 	memset(a2dp, 0, sizeof(*a2dp));
 
 	a2dp->refcnt = 1;
+	a2dp->seq_num = 1;
 
 	a2dp->state = BT_OPEN;
 
@@ -1053,6 +1212,9 @@ static inline void a2dp_free(snd_pcm_a2dp_t *a2dp)
 	if (a2dp->sk > fileno(stderr))
 		close(a2dp->sk);
 
+	if (a2dp->control_sk > fileno(stderr))
+		close(a2dp->control_sk);
+
 	sbc_finish(&a2dp->sbc);
 
 	free(a2dp);
@@ -1084,6 +1246,12 @@ static void a2dp_timer(snd_async_handler_t *async)
 	a2dp_unlock();
 }
 
+void sighand(int signo)
+{
+  printf("Thread in signal handler %d\n", signo);
+  return;
+}
+
 static void a2dp_init(void)
 {
 	snd_async_handler_t *async;
@@ -1093,6 +1261,13 @@ static void a2dp_init(void)
 	char timername[64];
 	int err, i;
 
+	// set up thread signal handler
+	memset(&actions, 0, sizeof(actions));
+	sigemptyset(&actions.sa_mask);
+	actions.sa_flags = 0;
+	actions.sa_handler = sighand;
+	sigaction(SIGALRM,&actions,NULL);
+
 	a2dp_lock();
 
 	for (i = 0; i < MAX_CONNECTIONS; i++)
@@ -1173,6 +1348,8 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a2dp)
 	snd_config_iterator_t i, next;
 	bdaddr_t src, dst;
 	int err, n, pos = -1, use_rfcomm = 0;
+	long bitpool = -1,subbnd = -1, blklen = -1;
+
 
 	DBG("name %s mode %d", name, mode);
 
@@ -1208,6 +1385,34 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a2dp)
 			continue;
 		}
 
+		if (!strcmp(id, "sbc_bitpool")) {
+			if ((err = snd_config_get_integer(n, &bitpool )) < 0 &&
+					( bitpool < 0 || 254 < bitpool))  {
+                                SNDERR("The field for bitpool must be an unsigned integer");
+                                return err;
+                        }
+			continue;
+		}
+
+		if (!strcmp(id, "sbc_blklen")) {
+			if ((err = snd_config_get_integer(n, &blklen)) < 0 &&
+					( blklen != 4 && blklen != 8 && blklen != 12 && blklen != 16)) {
+				SNDERR("The field for blocklen must be either 4, 8, 12 or 16");
+				return err;
+			}
+			continue;
+		}
+
+		if (!strcmp(id, "sbc_subband")) {
+			if ((err = snd_config_get_integer(n, &subbnd)) < 0 &&
+				 ( subbnd != 4 && subbnd != 8)) {
+				SNDERR("The field for subband must be either 4 or 8");
+				return err;
+			}
+			continue;
+		}
+
+
 		if (!strcmp(id, "use_rfcomm")) {
 			if ((err = snd_config_get_bool(n)) < 0) {
                                 SNDERR("The field use_rfcomm must be a boolean type");
@@ -1256,6 +1461,25 @@ SND_PCM_PLUGIN_DEFINE_FUNC(a2dp)
 		bacpy(&a2dp->dst, &dst);
 		a2dp->use_rfcomm = use_rfcomm;
 	}
+	a2dp->sbc.subbands = 8; // safe default
+	a2dp->sbc.blocks = 16; // safe default
+	a2dp->sbc.bitpool = 32; // recommended value 53, safe default is 32
+
+	if (bitpool != -1){
+		DBG("sbc : bitpool %d" , (int)bitpool);
+		a2dp->sbc.bitpool = (int)bitpool;
+	}
+
+	if (blklen != -1){
+		DBG("sbc : subbands %d" , (int)subbnd);
+		a2dp->sbc.subbands = (int)subbnd;
+	}
+
+	if (blklen != -1){
+		DBG("sbc : blklen %d" , (int)blklen);
+		a2dp->sbc.blocks = (int)blklen;
+	}
+
 
 	a2dp_unlock();
 
