@@ -703,12 +703,12 @@ int main(int argc, char *argv[])
 								fflush(stdout);
 							}
 							if (infobuf.mixer_volume[0] != last_volumes[0]) {
-								sprintf(line, "AT+VGS=%d\r",
+								sprintf(line, "\r\nAT+VGS=%d\r\n",
 									infobuf.mixer_volume[0]);
 								write(rd, line, strlen(line));
 							}
 							if (infobuf.mixer_volume[1] != last_volumes[1]) {
-								sprintf(line, "AT+VGM=%d\r",
+								sprintf(line, "\r\nAT+VGM=%d\r\n",
 									infobuf.mixer_volume[1]);
 								write(rd, line, strlen(line));
 							}
