@@ -90,7 +90,7 @@ MODULE_DESCRIPTION("Bluetooth SCO Headset Soundcard");
 MODULE_LICENSE("GPL");
 MODULE_SUPPORTED_DEVICE("{{ALSA,Bluetooth SCO Soundcard}}");
 
-static char *mod_revision = "$Revision: 1.12 $";
+static char *mod_revision = "$Revision: 1.13 $";
 
 static int index[SNDRV_CARDS] = {[0 ... (SNDRV_CARDS - 1)] = -2}; /* Exclude the first card */
 
@@ -1016,7 +1016,7 @@ static int snd_card_bt_sco_thread(void *data)
 #ifdef AUTO_FIXUP_BYTESHIFT
 			if (lostatcnt==len/2&&len>32) {
 				shift=!shift;
-				dprintk("Shift problem detected! Fixing to %d.\n",shift);
+				//dprintk("Shift problem detected! Fixing to %d.\n",shift);
 			}
 #endif
 #endif /* any of them */
